% Dependencies:
% Limb (class), 
classdef Robot < handle
    properties (SetAccess = private)
        name;
        limbs;  % Arrays of links
        T0;     % Transformation from inertia frame
        
        q;      % Symbolic generalized coordinates
        dq;     % Symbolic generalized coordinates derivative
        d2q;  	% Symbolic generalized coordinates derivative
        
        g;      % Gravity vector in global coordinates
        
        CoMlink;	% Virtual link from handhold to CoM
        
        functionsUpdated;
        
        Dfun;
        DfunTop;
        Cfun;
        CfunTop;
        Gfun;
        GfunTop;
        
        PhiVec;
        dPhiVec;
        d2PhiVec;
    end
    
    properties
        % Optimization properties
        kVal;
        tVal;
        theta;
        dtheta;
        d2theta;
        qVal;
        dqVal;
        d2qVal;
        tauVal;
    end
    
    properties (Dependent = true)
        n;      % Number of limbs on robot
        o;      % Origin of robot
        R;      % Rotation of robot
        
        r0;     % CoM location in inertia frame
        
        N;      % Number of generalized coordinates
    end
    
    methods
        % ========== Constructor ==========
        function obj = Robot(name, T0, q, dq, d2q, g)
            obj.functionsUpdated = 0;
            if nargin > 0
                obj.name = name;            
                obj.T0 = T0;
                obj.q = q;
                obj.dq = dq;
                obj.d2q = d2q;
                obj.g = g;
            end
            if nargin > 6
                error('Too many inputs');
            end
        end
        
        
        % ========== Get methods ==========
        function value = get.n(obj)
            value = length(obj.limbs);
        end
        
        function value = get.o(obj)
            value = obj.T0(1:3,4);
        end
        
        function value = get.R(obj)
            value = obj.T0(1:3,1:3);
        end
        
        function value = get.q(obj)
            value = obj.q;
        end
        
        function N = get.N(obj)
            N = length(obj.q);
        end
        
        function r0 = get.r0(obj)
            r0 = obj.CoMlink.r0;
        end
        
%         function link = getLink(obj, linkIndex)
%             ind = linkIndex;
%             for i=1:obj.n
%                 limb = obj.limbs(i);
%                 if limb.n > ind
%                     link = limb.links(ind);
%                     return
%                 else
%                     ind = ind - limb.n;
%                 end
%             end
%             error(['Link ', num2str(linkIndex), ' not found']);
%         end
        
        % ========== Other methods ==========
        function setSync(obj, PhiVec, dPhiVec, d2PhiVec)
            obj.PhiVec = PhiVec;
            obj.dPhiVec = dPhiVec;
            obj.d2PhiVec = d2PhiVec;
        end
        
        function setDynamics(obj, Dfun, Cfun, Gfun, DfunTop, CfunTop, GfunTop)
            obj.Dfun = Dfun;
            obj.Cfun = Cfun;
            obj.Gfun = Gfun;
            if nargin == 7
                obj.DfunTop = DfunTop;
                obj.CfunTop = CfunTop;
                obj.GfunTop = GfunTop;
            end
        end
        
        function origins = origins(obj)
            % Returns a matrix of origins of all links in the robot
            origins = [];
            for i=1:obj.n
                origins = [origins, NaN*ones(3,1), obj.limbs(i).origins];
            end
        end
        
        function massOrigins = massOrigins(obj)
            massOrigins = [];
            for i=1:obj.n
                massOrigins = [massOrigins, obj.limbs(i).massOrigins];
            end
        end
        
        function limb = addLimb(obj, name, T0)
            limb = Limb(obj.q, name, T0);
            obj.limbs = [obj.limbs, limb];
        end
        
        function plot(obj, q, framePlot, figTitle, d3Plot, mPlot, comPlot)
            if nargin < 3 || isempty(framePlot)
                framePlot = 0;
            end
            if nargin < 4 || isempty(figTitle)
                figTitle = '';
            end
            if nargin < 5 || isempty(d3Plot)
                d3Plot = 0;
            end
            if nargin < 6 || isempty(mPlot)
                mPlot = 0;
            end
            if nargin < 7 || isempty(comPlot)
                comPlot = 0;
            end
            
%             fun = matlabFunction(obj.origins, 'vars', {obj.q});
%             mfun = matlabFunction(obj.massOrigins, 'vars', {obj.q});
%             
%             oeval = fun(q);
%             moeval = mfun(q);
            oeval = eval(subs(obj.origins, {obj.q}, {q}));
            moeval = eval(subs(obj.massOrigins, {obj.q}, {q}));
            cmeval = eval(subs([obj.T0(1:3,4), obj.r0], {obj.q}, {q}));
            
%             plot3(oeval(1,:), oeval(2,:), oeval(3,:));
%             hold all
%             plot3(moeval(1,:), moeval(2,:), moeval(3,:), '*');
%             hold on
%             plot3(cmeval(1,:), cmeval(2,:), cmeval(3,:), 'r-*');
%             hold off
%             
%             xlabel('X'), ylabel('Y'), zlabel('Z'),
%             set(gca, 'DataAspectRatio', [1,1,1]);
%             axis([-0.1, 0.1, -0.5, 0.5, -1, 0.1]);
%             axis([-1,2,-1,1,-1,0.4]);
%             title(figTitle)
            p1 = {};
            p2 = {};
            p3 = {};
            if ~d3Plot
                p1 = {oeval(1,:), oeval(3,:), 'b'};
%                 plot(oeval(1,:), oeval(3,:));
%                 hold all
                if mPlot
                    p2 = {moeval(1,:), moeval(3,:), 'g*'};
%                     plot(moeval(1,:), moeval(3,:), '*');
                end
                if comPlot
                    p3 = {cmeval(1,:), cmeval(3,:), 'ro'};
%                     plot(cmeval(1,:), cmeval(3,:), 'ro');
                end
                plot(p1{:}, p2{:}, p3{:})
                axis([-1.5,1.5,-2,0.4]);
            else
                plot3(oeval(1,:), oeval(2,:), oeval(3,:));
                hold all
                if mPlot
                    plot3(moeval(1,:), moeval(2,:), moeval(3,:), '*');
                end
                if comPlot
                    plot3(cmeval(1,:), cmeval(2,:), cmeval(3,:), 'r-*');
                end
                axis([-1.5,1.5,-1,1,-2,0.4]);
            end
            
            xlabel('X'), ylabel('Y'), zlabel('Z')
            set(gca, 'DataAspectRatio', [1,1,1]);
            axis([-1.5,1.5,-2,0.4]);
            title(figTitle)
            
            if framePlot
                hold on
                for i=1:obj.n
                    for j=1:obj.limbs(i).n
                        drawframe(obj.limbs(i).links(j).T(q), 1/10);
                    end
                end
                hold off
            end
        end
        
        function jacobianSym(obj)
            for i=1:obj.n
                obj.limbs(i).jacobianSym();
            end
        end
        
        function calculateCoM(obj)
            obj.CoMlink = Link(obj.q, eye(4), obj.T0);
            for i = 1:length(obj.limbs)
                limb = obj.limbs(i);
                for j=1:length(limb.links)
                    link = limb.links(j);
                    if ~isempty(link.m)
                        r = obj.T0\[link.r0;1];
                        r = r(1:3);
                        obj.CoMlink.addMass(link.m, r, link.I0z);
                    end
                end
            end
        end
        
        function relabel(obj, newT0)
            % Specific for gorilla robot
            % Transformation from old T0 to new T0
            Tdiff = newT0/obj.T0;       
            obj.T0 = newT0;     
            
            % Reverse arm links
%             links = obj.limbs(1).links; % Arm links
%             for i=1:floor(length(links))
%                 tempLink = links(i);
%                 links(i) = links(end-i+1);
%                 links(end-i+1) = tempLink;
%             end
            
            % Change T0 of all links
            for i=1:obj.n
                limb = obj.limbs(i);
                limb.updateT0(Tdiff*limb.T0);
                for j=1:limb.n
                    link = limb.links(j);
                    link.updateT0(Tdiff*link.T0);
                end
            end
            
            % Gorilla spesifics
            T0_offset = zeros(4);
            T0_offset(2,4) = -newT0(2,4);
            for i=3:obj.limbs(1).n
                obj.limbs(1).links(i).updateT0(obj.limbs(1).links(i).T0 + 2*T0_offset);
            end
            obj.limbs(2).updateT0(obj.limbs(2).T0+T0_offset);
            for i=1:obj.limbs(2).n
                obj.limbs(2).links(i).updateT0(obj.limbs(2).links(i).T0 + T0_offset);
            end
                
        end

        % ======== Simulation methods ==========
        function [t, x] = simulateFull(obj, withTorque, tspan, x0, options)
            if nargin < 5
                options = [];
            end
            
            if withTorque
            	f = @obj.Ffull;
            else
                f = @(t,x) [x(end/2+1:end); - obj.Gfun(x(1:end/2)) - obj.Cfun(x(1:end/2), x(end/2+1:end))*x(end/2+1:end)];
            end
            M = @(t,x) blkdiag(eye(length(x)/2), obj.Dfun(x(1:end/2)));
            
            options = odeset(options, 'Mass', M);
            [t,x] = ode45(f, tspan, x0, options);
        end
        
        function M = animate(obj, t, qMat, fps, varargin)
            if length(t) < fps*t(end)
                disp(['Warning: not enough frames for smooth playback with ', num2str(fps), ' fps']);
            end
            
            currentframe = 0;
            for i=1:length(t)
                if t(i) > currentframe/fps
                    currentframe = currentframe + 1;
                    q = qMat(i,:)';
                    obj.plot(q, 0, ['t = ', num2str(t(i))], varargin{:});
%                     hold on
                    M(currentframe) = getframe;
                end
            end
            hold off
        end
        
        function [t,q,dq,d2q,tau] = simulateReduced(obj, k, tspan, x0, xend, options)
            if nargin < 6
                options = [];
            end
            
            N = size(k, 1)+1;

            
            theta_b = x0(1);
            theta_e = xend(1);
                       
            M = @(t, x) diag([1, alpha(x(1))]);
            f = @(t, x) [x(2); -beta(x(1))*x(2)^2 - gamma(x(1))];
                
            options = odeset(options, 'Mass', M);
            [t,x] = ode45(f, tspan, x0, options);
            
            dx = zeros(size(x));
            for i=1:length(t)
                dx(i,:) = M([], x(i,:))\f([],x(i,:));
            end
            
            theta = x(:,1);
            dtheta = x(:,2);
            d2theta = dx(:,2);
            
            Phi = obj.PhiVec(theta,k);
            dPhi = obj.dPhiVec(theta,k);
            d2Phi = obj.d2PhiVec(theta,k);

            q = Phi;
            dq = reshape(dPhi(:).*repmat(dtheta,obj.N,1), size(q));
            d2q = reshape(d2Phi(:).*repmat(dtheta.^2,obj.N,1) + dPhi(:).*repmat(d2theta,obj.N,1), size(q));

            tau = zeros(size(q));
            for i=2:length(theta)
                qi = q(i,:)';
                dqi = dq(i,:)';
                d2qi = d2q(i,:)';
                tau(i,:) = obj.Dfun(qi)*d2qi + obj.Cfun(qi, dqi)*dqi + obj.Gfun(qi);
            end
            
            obj.tVal = t;
            obj.kVal = k;
            obj.qVal = q;
            obj.dqVal = dq;
            obj.d2qVal = d2q;
            obj.tauVal = tau;
            
            function alpha = alpha(theta)
                alpha = obj.DfunTop(obj.PhiVec(theta, k, theta_b, theta_e)')*obj.dPhiVec(theta, k, theta_b, theta_e)';
            end
            function beta = beta(theta)
                Phi = obj.PhiVec(theta,k,theta_b,theta_e)';
                dPhi = obj.dPhiVec(theta,k,theta_b,theta_e)';
                d2Phi = obj.d2PhiVec(theta,k,theta_b,theta_e)';

                beta = obj.CfunTop(Phi, dPhi)*dPhi + obj.DfunTop(Phi)*d2Phi;
            end
            function gamma = gamma(theta)
                gamma = obj.GfunTop(obj.PhiVec(theta,k,theta_b,theta_e)');
            end
        end
        
%         function [q,dq,d2q,tau] = expandReduced(obj, k, theta, dtheta, d2theta)
%             Phi = obj.PhiVec(k,theta);
%             dPhi = obj.dPhiVec(k,theta);
%             d2Phi = obj.d2PhiVec(k,theta);
% 
%             q = Phi;
%             dq = reshape(dPhi(:).*repmat(dtheta,obj.N,1), size(q));
%             d2q = reshape(d2Phi(:).*repmat(dtheta.^2,obj.N,1) + dPhi(:).*repmat(d2theta,obj.N,1), size(q));
% 
%             if nargout > 3
%                 tau = obj.torque(q,dq,d2q);
%             end
%         end
    end
end