% Dependencies:
% Link (class)
classdef Limb < handle
    properties (SetAccess = private)
        name;
        links;  % Array of links
        T0;     % Transformation from inertia frame to first link
        q;      % Symbolic generalized coordinates
    end
    
    properties (Dependent = true)
        n;      % Number of links in limb
        o;      % Origin of limb
        R;      % Rotation of limb
    end
    
    methods
        % ========== Constructor ==========
        function obj = Limb(q, name, T0)
            if nargin > 0
                obj.q = q;
                obj.name = name;
                obj.T0 = T0;     % Transformation from inertia frame
            end
        end
        
        % ========== Get methods ==========
        function value = get.n(obj)
            value = length(obj.links);
        end
        
        function value = get.o(obj)
            value = obj.T0(1:3,4);
        end
        
        function value = get.R(obj)
            value = obj.T0(1:3,1:3);
        end
        
        % ========== Other methods ==========
        function obj = updateT0(obj, T0);
            obj.T0 = T0;
        end
        
        function origins = origins(obj)
            % Returns a matrix of all origins of the limb
            origins = sym(zeros(3,obj.n+1));
            origins(:,1) = obj.o;
            for i=1:obj.n
                origins(:,i+1) = obj.links(i).o;
            end
        end
        
        function massOrigins = massOrigins(obj)
            massOrigins = [];
            for i=1:obj.n
                if ~isempty(obj.links(i).r)
                    massOrigins = [massOrigins, obj.links(i).r0vec];
                end
            end
        end
        
        function link = addLink(obj, varargin)              
            if obj.n == 0
                T0prev = obj.T0;
            else
                T0prev = obj.links(end).T0;
            end
            link = Link(obj.q, DH(varargin{:}), T0prev);
            obj.links = [obj.links, link];
        end              
        
        function addMass(obj, varargin)
            % Sets the mass, mass center location and inertia matrix of the
            % last link
            obj.links(end).addMass(varargin{:});
        end
        
        function jacobianSym(obj)
            for i=1:obj.n
                obj.links(i).jacobianSym();
            end
        end
        
    end
end