% Dependencies:
% InertiaTensor (function), DH (function)
classdef Link < handle
    properties (SetAccess = private)
        Tl;     % Transformation from previous link frame (limb frame if first link)
        T0;     % Transformation from inertia frame
        rho;    % 1 if link is rotational, 0 otherwise
        
        rvec    % Location of mass centers in local frame (3xK matrix)
        
        m;      % Sum of masses
        r;      % Location of CoM in local frame (3x1 vector)
        I;      % Inertia about CoM (3x3 matrix)
        
        q;      % Symbolic generalized coordinates
        qSym;   % Generalized coordinate assosiated with link
        
        Jvel;
        Jrot;
    end
    
    properties (Dependent = true)
        o;      % Origin in inertia frame
        R;      % Rotation from inertia frame
        r0vec;  % Location of all masses in link in inertia frame
        r0;     % Location of CoM in inertia frame
        I0;     % Moment of inertia in inertia frame
        I0z;    % Moment of inertia in inertia frame at zero config
        
        Rz;     % Rotation matrix evaluated at the zero config
        Tz;     % Transformation matrix evaluated at the zero config
        Tlz;    % Local transformation evaluated at the zero config
        
    end
    
    methods
        % ========== Constructor ==========
        function obj = Link(q, Tl, T0prev, rho)
            if nargin > 0
                if nargin < 7
                    rho = 1;
                end
                
                obj.q = q;
                obj.rho = rho;
                obj.Tl = Tl;                % Local transformation
                obj.T0 = T0prev*obj.Tl;     % Global transformation
            end
        end
        
        function new = copy(this)
            new = eval(class(this));
            mo = eval(['?',class(this)]);
            ndp = findobj([mo.Properties{:}],'Dependent',false);
            for i=1:length(ndp)
                new.(ndp(i).Name) = this.(ndp(i).Name);
            end
        end        
        
        % ========== Get methods ==========
        function value = get.o(obj)
            value = obj.T0(1:3,4);
        end
        
        function value = get.R(obj)
            value = obj.T0(1:3,1:3);
        end
        
        function value = get.I0(obj)
            value = obj.R*obj.I*transpose(obj.R);
        end
        
        function value = get.I0z(obj)
            value = obj.Rz*obj.I*transpose(obj.Rz);
        end
        
        function value = get.Rz(obj)
            value = subs(obj.R, {obj.q}, zeros(size(obj.q)));
%             value = eval(subs(obj.R, {obj.q}, zeros(size(obj.q))));
        end
        
        function value = get.Tz(obj)
            value = eval(subs(obj.T0, {obj.q}, zeros(size(obj.q))));
        end
        
        function value = get.Tlz(obj)
            value = eval(subs(obj.Tl, {obj.q}, zeros(size(obj.q))));
        end
        
        function r0vec = get.r0vec(obj)
            k = size(obj.rvec,2);
            rh = [obj.rvec;ones(1,k)];
            rh0 = obj.T0*rh;
            r0vec = rh0(1:3,:);
        end
        
        function r0 = get.r0(obj)
            rh = [obj.r;1];
            rh0 = obj.T0*rh;
            r0 = rh0(1:3);
        end
        
        % ========== Other methods ========== 
        function obj = updateT0(obj, T0)
            obj.T0 = T0;
        end
        
        function addMass(obj, m, r, I)
            if ~isscalar(m)
                error('mIn should be a scalar');
            end
            if (size(r,1) ~= 3 || size(r,2) ~= 1)
                error('rIn should be a 3x1 vector');
            end
            if (size(I,1) ~= 3 || size(I,2) ~= 3)
                error('IIn should be a 3x3 matrix');
            end
            
            % Append to mass location matrix
            obj.rvec = [obj.rvec, r];
            
            obj.sumMass(m, r, I);
        end
        
        function sumMass(obj, mIn, rIn, IIn)
            % Rotate I to local frame (should be given in global
            % zero-config coordinates)
            IIn = obj.Rz'*IIn*obj.Rz;
          
            if isempty(obj.m)
                obj.m = mIn;
                obj.r = rIn;
                obj.I = IIn;
            else
                mOld = obj.m;
                rOld = obj.r;
                IOld = obj.I;

                mNew = mOld+mIn;                        % sum of masses
                rNew = 1/mNew*(mOld*rOld + mIn*rIn);    % local CoM location

                dr1 = rOld - rNew;
                dr2 = rIn - rNew;
                INew = IOld + InertiaTensor(mOld, dr1) + IIn + InertiaTensor(mIn, dr2);

                obj.m = mNew;
                obj.r = rNew;
                obj.I = INew;
            end
        end
            
        function jacobianSym(obj)
            if ~isempty(obj.m)
                q = obj.q;
                obj.Jvel = jacobian(obj.r0, q);
                obj.Jvel = simplify(obj.Jvel);
                
                R = obj.R;
                obj.Jrot = [jacobian(R(3,1),q)*R(2,1) + jacobian(R(3,2),q)*R(2,2) + jacobian(R(3,3),q)*R(2,3);
                            jacobian(R(1,1),q)*R(3,1) + jacobian(R(1,2),q)*R(3,2) + jacobian(R(1,3),q)*R(3,3);
                            jacobian(R(2,1),q)*R(1,1) + jacobian(R(2,2),q)*R(1,2) + jacobian(R(2,3),q)*R(1,3)];
                obj.Jrot = simplify(obj.Jrot);
            end
        end
        
        function value = T(obj, q)
            if size(q) ~= size(obj.q)
                error('q should be the same length as the generalized coordinates');
            end
            value = eval(subs(obj.T0, {obj.q}, {q}));
        end
    end
end

