function [D_hat, C_hat, G_hat, B_hat] = reduceDynamics(D,C,G,B,fqr,F,q,dq,d2q,rset)
% Reduces the coordinates given by qr when the system
%   D(q)*d2q + C(q,dq)*dq + G(q) = B(q)*u
% is subject to the constraint F s.t. F(q) = 0.
% rset are the indices of the redundant coordinates

global dirs

if (size(fqr, 1) ~= size(F,1) || size(fqr,2) ~= size(F,2))
    error('The number of constraints must equal the number of coordinates to be eliminated');
end

n = length(q);
r = size(F,1);
m = n-r;

if nargin < 10
    rset = m+1:n;
end

nset = 1:n;
mset = setdiff(nset, rset);

% Split input vectors
qm = q(mset);
dqm = dq(mset);
d2qm = d2q(mset);

qr = q(rset);
dqr = dq(rset);
d2qr = d2q(rset);

% Calculate jacobian
J = jacobian(F,q);      % J*dq = 0
dJ = sym(zeros(size(J)));
for i=1:r
    dJ(i,:) = dq'*jacobian(J(i,:),q);
end

Jm = J(:,mset);
Jr = J(:,rset);
dJm = dJ(:,mset);
dJr = dJ(:,rset);

% EOM with contact forces
Dt = D(mset,:);
Db = D(rset,:);
Ct = C(mset,:);
Cb = C(rset,:);
Gt = G(mset,:);
Gb = G(rset,:);
Bt = B(mset,:);
Bb = B(rset,:);

D_tilde = Dt - Jm'/(Jr')*Db;
C_tilde = Ct - Jm'/(Jr')*Cb;
G_tilde = Gt - Jm'/(Jr')*Gb;
B_tilde = Bt - Jm'/(Jr')*Bb;


% Eliminate redundant coordinates
Gamma = simplify(-Jr\Jm);
Gamma = subs(Gamma,qr,fqr);
fdqr = Gamma*dqm;
Delta = simplify(-Jr\(dJm - dJr/Jr*Jm));
Delta = subs(Delta, [qr;dqr], [fqr;fdqr]);

D_hat = D(mset,mset) + D(mset,rset)*Gamma - Jm'/Jr'*(D(rset,mset)+D(rset,rset)*Gamma);
C_hat = C(mset,mset) + C(mset,rset)*Gamma + D(mset,rset)*Delta - Jm'/Jr'*(C(rset,mset) + C(rset,rset)*Gamma + D(rset,rset)*Delta);
G_hat = G(mset) - Jm'/Jr'*G(rset);
B_hat = B(mset,:) - Jm'/Jr'*B(rset,:);

D_hat = subs(D_hat, [qr;dqr], [fqr;fdqr]);
C_hat = subs(C_hat, [qr;dqr], [fqr;fdqr]);
G_hat = subs(G_hat, [qr;dqr], [fqr;fdqr]);
B_hat = subs(B_hat, [qr;dqr], [fqr;fdqr]);


folder = [dirs.gorilla, filesep, 'SyncronizationFunctions'];
qr_fun = matlabFunction(fqr, 'vars', {qm}, 'file', [folder, filesep 'fqrFun']);
dqr_fun = matlabFunction(Gamma*dqm, 'vars', {qm,dqm}, 'file', [folder, filesep, 'dfqrFun']);
d2qr_fun = matlabFunction(Gamma*d2qm+Delta*dqm, 'vars', {qm,dqm,d2qm}, 'file', [folder, filesep, 'd2fqrFun']);

end