function B = bezier(t_in, P, t0, tend)
% Bezier curve of time t_in with control points given in P.
%   Syntax:
%       B = bezier(t, P) returns a matrix of values corresponding to the
%           bezier curves with control points P. Input t must be a vector.
%
%       B = bezier(t, P, t0, tend) returns a vector containing the values of the
%           bezier curves with control points P spanning from t0 to tend
%           evaluated at t.
%       
%       P should be a m-by-n matrix where m is the number of bezier curves
%       to be produced and n is the number of control points in each curve.
%
%       If t is a vector, t0 and tend are ignored and the first
%       syntax is used.
%
%   Note:
%       B is such that time runs along the dimension indicated by
%       the time input. I.e. if t is a column vector, time will run along
%       dimension 1. If t is a row vector, time will run along dimension 2.
%       If t is scalar, B will be a column vector.


% Input handling
if isscalar(t_in) && nargin < 4
    error('t_min and t_max must be provided when t is scalar');
elseif ~isscalar(t_in) && ~isvector(t_in)
    error('Input t must be scalar or a vector');
elseif ~isscalar(t_in) && isvector(t_in)
    t0 = t_in(1);
    tend = t_in(end);
    if nargin > 2
        warning('Ignoring provided t_min and t_max (t_in is a vector)');
    end
end

if iscolumn(t_in)
    t = t_in';
else
    t = t_in;
end

% Calculations
m = size(P,1);                  % Number of independent Bzier curves to be created
n = size(P,2);                  % Order of the Bzier curve
t = (t-t0)/(tend-t0);           % Remap time to [0,1]

pascal_mat = pascal(n);
ncvec = pascal_mat(sub2ind([n,n],n:-1:1,1:n));                      % Contains nchoosek(n,i) for i=0:n
tmat = repmat(t,n-1,1);                                             % All rows contains identical copies of t
t1vec = [ones(size(t)); cumprod(tmat,1)];                           % Row i contains t.^(i-1)
t2vec = flipud([ones(size(t)); cumprod(ones(size(tmat))-tmat,1)]);  % Row i (from the bottom) contains (1-t).^(i-1)

B = (repmat(ncvec,m,1).*P)*(t1vec.*t2vec);

% Output handling
if ~isscalar(t_in) && iscolumn(t_in)
    B = B';         % Transverse B such that time runs along the same dimension as in t
end

end


